from _pytest.capture import CaptureFixture
from packaging.version import Version

from piplexed.display.print_tree import print_installed_tree
from piplexed.display.print_tree import print_outdated_tree
from piplexed.venvs.pipx_venvs import PackageInfo


def test_print_installed_tree(capsys: CaptureFixture[str]):
    packages = [
        PackageInfo("A", Version("1.0.1"), python="3.11.3"),
        PackageInfo("B", Version("2.2.1"), python="3.11.3"),
    ]
    print_installed_tree(packages, "pipx")

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("PIPX Packages")
    assert output[1] == "├── A"
    assert output[2] == "│   ├── version - 1.0.1"
    assert output[3] == "│   └── python - 3.11.3"


def test_print_outdated_tree(capsys: CaptureFixture[str]):
    packages = [
        PackageInfo("B", Version("1.5.0"), python="3.11.3", latest_pypi_version=Version("1.6.0")),
        PackageInfo("C", Version("2.6.0"), python="3.11.3", latest_pypi_version=Version("3.0a.0")),
    ]

    print_outdated_tree(packages, "pipx")

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    print(out)
    assert output[0] == "PIPX Outdated Packages"
    assert output[1] == "├── B"
    assert output[2] == "│   ├── installed version - 1.5.0"
    assert output[3] == "│   └── PyPI version - 1.6.0"
    assert output[4] == "└── C"
    assert output[5] == "├── installed version - 2.6.0"
    assert output[6] == "└── PyPI version - 3.0a0 ⚠"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from packaging.utils import NormalizedName
from packaging.version import Version


class ToolType(str, Enum):
    PIPX = "pipx"
    UV = "uv"
    ALL = "all"


@dataclass
class PackageInfo:
    name: NormalizedName
    version: Version
    python: str | None = None
    latest_pypi_version: Version | None = None
    tool: ToolType = ToolType.PIPX

    def newer_pypi_version(self) -> bool:
        if self.latest_pypi_version is not None:
            return self.latest_pypi_version > self.version
        else:
            return False


__all__ = [
    "ToolType",
    "PackageInfo",
]



from __future__ import annotations

import typer

import piplexed
from piplexed.app import print_installed_tools
from piplexed.app import print_outdated_tools
from piplexed.venvs import ToolType

app = typer.Typer()


@app.command()
def version() -> None:
    print(piplexed.__version__)


@app.command()
def list(
    outdated: bool = typer.Option(
        False, "--outdated", "-O", help="Find installed packages/tools with newer versions on PyPI"
    ),
    is_prerelease: bool = typer.Option(
        False,
        "--pre",
        "-P",
        help="Include pre and dev releases in latest pypi version search",
    ),
    tree: bool = typer.Option(False, "--tree", "-T", help="print output as a tree (default is table)"),
    tool: ToolType = typer.Option(
        "pipx", "--tool", help="choose tool packages were installed with, 'pipx', 'uv', 'all"
    ),
) -> None:
    if outdated:
        print_outdated_tools(is_prerelease=is_prerelease, tree=tree, tool=tool)
    else:
        print_installed_tools(tree=tree, tool=tool)


if __name__ == "__main__":
    raise SystemExit(app)



from __future__ import annotations

import platform
import shutil
import subprocess
from pathlib import Path

from packaging.utils import canonicalize_name
from packaging.version import Version

from piplexed.venvs import PackageInfo
from piplexed.venvs import ToolType


def find_uv() -> tuple[bool, str]:
    uv_path = shutil.which("uv")
    return uv_path is not None, uv_path or "uv"


def find_uv_tool_dir(uv_path: str) -> str | None:
    # uv tool dir will return a tool directory even if one does not exist
    ret = subprocess.run([uv_path, "tool", "dir"], check=False, text=True, capture_output=True)  # noqa: S603

    if ret.returncode == 0:
        tool_dir = ret.stdout.strip()
        return tool_dir if Path(tool_dir).exists() else None

    return None


HAS_UV, UV_PATH = find_uv()

# If we can't find uv then there is no tool directory
UV_TOOL_DIR = find_uv_tool_dir(UV_PATH) if HAS_UV else None
UV_VENV_DIR = Path(UV_TOOL_DIR) if UV_TOOL_DIR is not None else None
OS_SYSTEM = platform.system()


def get_python_path(tool_venv_dir: Path, op_sys: str) -> Path:
    if op_sys == "Windows":
        return tool_venv_dir / "Scripts" / "python.exe"

    return tool_venv_dir / "bin" / "python"


def uv_tool_version(tool_venv_dir: Path, op_sys: str) -> str:
    python_path = get_python_path(tool_venv_dir, op_sys)
    script = f"from importlib.metadata import version; print(version('{tool_venv_dir.name}'))"
    ret = subprocess.run([python_path, "-c", script], check=False, text=True, capture_output=True)  # noqa: S603

    return ret.stdout.strip()


def installed_uv_tools(tool_dir: Path | None = UV_VENV_DIR, op_sys: str = OS_SYSTEM) -> list[PackageInfo]:
    def run_subprocess(cmd: str) -> str:
        try:
            result = subprocess.run([python_path, "-c", cmd], capture_output=True, text=True, check=True)  # noqa: S603
            return result.stdout.strip()
        except subprocess.CalledProcessError as e:
            return f"Error: {e.stderr.strip()}"

    uv_packages: list[PackageInfo] = []

    if tool_dir is None:
        return uv_packages

    for env in tool_dir.iterdir():
        if not env.name.startswith("."):
            python_path = get_python_path(env, op_sys)

            package_version_cmd = f"from importlib.metadata import version; print(version('{env.name}'))"
            package_version = run_subprocess(package_version_cmd)

            python_version_cmd = "import sys; print(sys.version.split()[0])"
            python_version = run_subprocess(python_version_cmd)

            uv_data = PackageInfo(
                name=canonicalize_name(env.name),
                version=Version(package_version),
                python=python_version,
                tool=ToolType.UV,
            )

            uv_packages.append(uv_data)
    return uv_packages



from unittest.mock import patch

from typer.testing import CliRunner

from piplexed.cli import app
from piplexed.version import VERSION

runner = CliRunner()


def test_cli_list():
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0
    assert result.stdout.strip() == VERSION


def mock_installed_output(tree, tool):
    print(f"tree:{tree} tool:{tool.value}")


def mock_outdated_output(is_prerelease, tree, tool):
    print(f"is_prerelease:{is_prerelease} tree:{tree} tool:{tool.value}")


@patch("piplexed.cli.print_installed_tools", side_effect=mock_installed_output)
def test_list_cmd_defaults(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list"])
    assert result.exit_code == 0
    assert "tree:False tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_installed_tools", side_effect=mock_installed_output)
def test_list_cmd_uv_default(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--tool", "uv"])
    assert result.exit_code == 0
    assert "tree:False tool:uv" in result.stdout.strip()


@patch("piplexed.cli.print_installed_tools", side_effect=mock_installed_output)
def test_list_cmd_pipx_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--tree"])
    assert result.exit_code == 0
    assert "tree:True tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_installed_tools", side_effect=mock_installed_output)
def test_list_cmd_uv_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--tree", "--tool", "uv"])
    assert result.exit_code == 0
    assert "tree:True tool:uv" in result.stdout.strip()


def test_invalid_tool_option():
    result = runner.invoke(app, ["list", "--tool", "jeff"])
    assert result.exit_code != 0


@patch("piplexed.cli.print_installed_tools", side_effect=mock_installed_output)
def test_use_of_prerelease_has_no_effect_on_installed(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--pre"])
    assert result.exit_code == 0
    assert "tree:False tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_defaults(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated"])
    assert result.exit_code == 0
    assert "is_prerelease:False tree:False tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_pipx_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--tree"])
    assert result.exit_code == 0
    assert "is_prerelease:False tree:True tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_pipx_prerelease(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--pre"])
    assert result.exit_code == 0
    assert "is_prerelease:True tree:False tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_pipx_prerelease_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--pre", "--tree"])
    assert result.exit_code == 0
    assert "is_prerelease:True tree:True tool:pipx" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_uv_defaults(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--tool", "uv"])
    assert result.exit_code == 0
    assert "is_prerelease:False tree:False tool:uv" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_uv_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--tree", "--tool", "uv"])
    assert result.exit_code == 0
    assert "is_prerelease:False tree:True tool:uv" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_uv_prerelease(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--pre", "--tool", "uv"])
    assert result.exit_code == 0
    assert "is_prerelease:True tree:False tool:uv" in result.stdout.strip()


@patch("piplexed.cli.print_outdated_tools", side_effect=mock_outdated_output)
def test_list_cmd_outdated_uv_prerelease_tree(installed_package):
    installed_package.return_value = None
    result = runner.invoke(app, ["list", "--outdated", "--pre", "--tree", "--tool", "uv"])
    assert result.exit_code == 0
    assert "is_prerelease:True tree:True tool:uv" in result.stdout.strip()



from __future__ import annotations

import warnings
from os import getenv

from rich.console import Console
from rich.panel import Panel

TERMINAL_WIDTH = getenv("TERMINAL_WIDTH")


def future_deprecation_warning(*, reason: str, replacement: str | None, deprecation_version: str) -> None:
    """Helper function to more cleanly inform users of upcoming deprecations

    Parameters
    ----------
    reason : str
        An explanation about why this functionality has been deprecated.
    replacement : str | None
        Suggested alternative functionality to use
    deprecation_version : str
        The version of piplexed that will contain this change.
    """

    if replacement is None:
        replacement = ""

    message_parts = (reason, f"This will happen in {deprecation_version}", replacement)

    message = ". ".join(message_parts)

    warnings.warn(message.rstrip(), stacklevel=2, category=FutureWarning)


def _get_rich_console(stderr: bool = False) -> Console:  # noqa: FBT001, FBT002
    return Console(
        color_system="auto",
        width=int(TERMINAL_WIDTH) if TERMINAL_WIDTH else None,
        stderr=stderr,
    )


def rich_format_error(msg: str) -> None:
    console = _get_rich_console(stderr=True)
    console.print(Panel(msg, border_style="red", title="Error", title_align="left", width=70))


def rich_format_info(msg: str) -> None:
    console = _get_rich_console()
    console.print(Panel(msg, border_style="orange1", title="[blue]Info[/blue]", title_align="left", width=70))
